import Regexp ;


//! A ClassNavigator object gives you information about the classes which another inherits from, the methods of a class, etc...
class ClassNavigator 
{
  private master mst ;
  //! This regexp is used to clean the name of a class, which pike gives us with lots of strange characters.
  
  private SimpleRegexp re ;
  //! This regexp cleans things like ___Math.IMatrix().
  
  private SimpleRegexp re2 ;
  void create()
  {
    mst = master() ;
    re = SimpleRegexp(".+\\(\\)->(.+)\\(\\).*");
    re2 = SimpleRegexp("___(.+)\\(\\)");
  }
  
  //! Returns an array of strings, all of them naming a method or variable of the class to which a belongs.
  array(string) methodsAndVariablesOfClass( mixed a )
  {
    return indices(a);
  }
  
  //! Returns the string which names the class a .
  string nameOfClass( mixed a )
  {
    string res =  mst->describe_object(a);
    if ( re->match(res) ) 
      {
	return (re->split(res))[0] ;
      }
    else if ( re2->match(res) )
      {
	return (re2->split(res))[0] ;
      }
    else
      {
	return res ;
      }
  }
  
  //! Returns an array with all the classes which a DIRECTLY inherits from
  array(string) classInheritsFrom ( mixed a )
  {
    array aux =  Program.inherit_list(a) ;
    array resultat = ({});
    foreach( aux, program item )
      {
	string res =  mst->describe_module(item)  ;
	if ( re->match(res) )
	  {
	    res = (re->split(res))[0] ;
	  }
	else if ( re2->match(res) )
	  {
	    res = (re2->split(res))[0] ;
	  }
	resultat +=({ res }) ;
      }
    return resultat ;
  }
  
  //! Returns 1 if this class doesn't inherit from another.
  int classDoesNotInherit( mixed a )
  {
    array res = classInheritsFrom(a) ;
    if ( sizeof(res) == 0  )
      {
	return 1 ;
      }
    else 
      {
	return 0 ;
      }
  }
  
  //!Returns all the classes from the inheritance tree of a .
  array(string) classInheritanceTree( mixed a )
  {
    array aux1 = Program.inherit_list(a) ;
    array aux = ({});
    if ( sizeof(aux1) == 0 )
      {
	return 0 ;
      }
    if ( sizeof(aux1) > 0 )
      {
	aux = Program.inherit_tree(aux1[0]);
      }
    array resultat = ({});
    array aux2 = ({});
    array error = ({}) ;
    int seguir = 1 ;
    while(sizeof(aux) > 1 )
      {
	aux2 += ({aux[0] }) ;
	aux = aux[1];
      }
    aux2 += ({aux[0] }) ;
	   
    foreach( aux2, mixed item )
      {
	string p = mst->describe_module(item) ;
	if ( re->match(p) )
	  {
	    p = (re->split(p))[0] ;
	  }
	else if ( re2->match(p))
	  {
	    p = (re2->split(p))[0] ;
	  }
	resultat += ({ p }) ;
      }
    return resultat ;
  }
}


//These are 3 classes used as an example about how this module work.

class a {
  
  mixed variable_from_class_a ;
  void method_of_class_a () ;
} ;
class b { inherit a ; 
  void method_of_class_b () ;
} ;
class c { inherit b ; 
  void method_of_class_c () ;
} ;


int main ()
{
  ClassNavigator cl = ClassNavigator();
  c example = c () ;
  array r = cl->methodsAndVariablesOfClass(example );
  write("*** We have an instance of class " + cl->nameOfClass(example) + " ***\n" );
  write("--- Its methods and variables are : ---\n");
  foreach( r , string item )
    {
      write( item + "\n" ) ;
    }
  r = cl->classInheritsFrom( example ) ;
  write("--- Its direct superclasses are ---\n");
  if ( r )
    {
      foreach( r, string item )
	{
	  write(item + "\n");
	}
    }
  write("--- Its complete inheritance tree is: ---\n");
  r = cl->classInheritanceTree(example);
  if ( r )
    {
      foreach( r, string item )
	{
	  write(item + "\n");
	}
    }
  
  Math.IMatrix im = Math.IMatrix(2,3);
  write("*** We also have an instance of class " + cl->nameOfClass(im) + " ***\n" );
  r = cl->methodsAndVariablesOfClass(im );
  write("--- Its methods and variables are : ---\n");
  foreach( r , string item )
    {
      write( item + "\n" ) ;
    }
  r = cl->classInheritsFrom( im ) ;
  write("--- Its direct superclasses are ---\n");
  if ( r )
    {
      foreach( r, string item )
	{
	  
	  write(item + "\n");
	}
    }
  write("--- Its complete inheritance tree is: ---\n");
  r = cl->classInheritanceTree(im);
  if ( r )
    {
      foreach( r, string item )
	{
	  write(item + "\n");
	}
    }

  write("Is this a class that does not inherit from another?\n");
  if ( cl->classDoesNotInherit(im) )
    {
      write("Yes!\n");
    }
  else
    {
      write("No\n");
    }

  return 0 ;
}


